import fs from 'node:fs'
import path from 'node:path'
import { fileURLToPath } from 'node:url'
import esbuild from 'esbuild'

const __dirname = path.dirname(fileURLToPath(import.meta.url))

async function buildModule(inputPath, outputPath) {
  try {
    const result = await esbuild.build({
      entryPoints: [inputPath],
      bundle: true,
      write: false,
      format: 'cjs',
      platform: 'node',
      external: ['react', 'react-native'],
      define: {
        'process.env.NODE_ENV': '"production"',
      },
    })

    if (inputPath.includes('/themes/')) {
      const themeContent = result.outputFiles[0].text
      const match = themeContent.match(/JSON\.parse\('(.+?)'\)/)
      if (match) {
        fs.writeFileSync(outputPath, `module.exports = ${match[1]}`)
      }
      else {
        throw new Error('Could not find theme data in bundle')
      }
    }
    else {
      fs.writeFileSync(outputPath, result.outputFiles[0].text)
    }

    return true
  }
  catch (error) {
    console.error(`Failed to build ${inputPath}:`, error)
    return false
  }
}

const shikiDir = path.join(__dirname, '../node_modules/shiki/dist')
const langFiles = fs
  .readdirSync(path.join(shikiDir, 'langs'))
  .filter(file => file.endsWith('.mjs'))
  .map(file => ({
    input: `shiki/dist/langs/${file}`,
    output: `dist/cache/langs/${file.replace('.mjs', '.js')}`,
  }))

const themeFiles = fs
  .readdirSync(path.join(shikiDir, 'themes'))
  .filter(file => file.endsWith('.mjs'))
  .map(file => ({
    input: `shiki/dist/themes/${file}`,
    output: `dist/cache/themes/${file.replace('.mjs', '.js')}`,
  }))

const coreModules = [
  {
    input: 'shiki/dist/core.mjs',
    output: 'dist/cache/core/core.js',
  },
  {
    input: 'shiki/dist/engine-javascript.mjs',
    output: 'dist/cache/core/engine-javascript.js',
  },
  ...langFiles,
  ...themeFiles,
]

for (const module of coreModules) {
  const inputPath = path.join(__dirname, '../node_modules', module.input)
  const outputPath = path.join(__dirname, '..', module.output)

  const dir = path.dirname(outputPath)
  if (!fs.existsSync(dir))
    fs.mkdirSync(dir, { recursive: true })

  await buildModule(inputPath, outputPath)
  console.log(`Built ${path.basename(module.input)} -> ${path.basename(module.output)}`)
}

export async function buildOnDemand(type, name) {
  const inputPath = path.join(__dirname, `../node_modules/shiki/dist/${type}/${name}.mjs`)
  const outputPath = path.join(__dirname, `../dist/cache/${type}/${name}.js`)

  if (await buildModule(inputPath, outputPath)) {
    const manifestPath = path.join(__dirname, '../dist/cache/manifest.json')
    const manifest = JSON.parse(fs.readFileSync(manifestPath, 'utf8'))
    manifest[type][name] = { built: Date.now() }
    fs.writeFileSync(manifestPath, JSON.stringify(manifest, null, 2))
    return outputPath
  }
  return null
}
