import type { LanguageInput, ThemedToken, ThemeInput } from 'shiki'
import type { TokenType } from '../../types'
import { createHighlighterCore } from '../../../dist/cache/core/core.js'
import { createJavaScriptRegexEngine } from '../../../dist/cache/core/engine-javascript.js'

export type SupportedLanguage = 'css' | 'html' | 'javascript' | 'json' | 'php' | 'typescript'
export type SupportedTheme = 'github-dark' | 'github-light'

export interface HighlighterState {
  readonly highlighter: Awaited<ReturnType<typeof createHighlighterCore>>
  readonly loadedLanguages: ReadonlySet<string>
  readonly loadedThemes: ReadonlySet<string>
}

let state: HighlighterState | null = null

function getState(): HighlighterState | null {
  return state
}

function setState(newState: HighlighterState): void {
  state = newState
}

async function importLanguage(lang: string): Promise<LanguageInput> {
  try {
    const module = await import(`../dist/cache/langs/${lang}.js`)
    return module.default
  }
  catch (error) {
    console.error(`Failed to load language '${lang}':`, error)
    throw new Error(`Language '${lang}' not found in cache`)
  }
}

async function importTheme(theme: string): Promise<ThemeInput> {
  try {
    const module = await import(`../dist/cache/themes/${theme}.js`)
    return module.default
  }
  catch (error) {
    console.error(`Failed to load theme '${theme}':`, error)
    throw new Error(`Theme '${theme}' not found in cache`)
  }
}

async function initializeHighlighter(langs: string[], themes: string[]): Promise<Awaited<ReturnType<typeof createHighlighterCore>>> {
  const [langModules, themeModules] = await Promise.all([Promise.all(langs.map(importLanguage)), Promise.all(themes.map(importTheme))])

  return createHighlighterCore({
    engine: createJavaScriptRegexEngine(),
    langs: langModules,
    themes: themeModules,
  })
}

async function loadAdditionalResources(currentState: HighlighterState, langs: string[], themes: string[]): Promise<HighlighterState> {
  const newLangs = langs.filter(lang => !currentState.loadedLanguages.has(lang))
  const newThemes = themes.filter(theme => !currentState.loadedThemes.has(theme))

  if (newLangs.length === 0 && newThemes.length === 0) {
    return currentState
  }

  await Promise.all([
    ...newLangs.map(async (lang) => {
      const langModule = await importLanguage(lang)
      await currentState.highlighter.loadLanguage(langModule)
    }),
    ...newThemes.map(async (theme) => {
      const themeModule = await importTheme(theme)
      await currentState.highlighter.loadTheme(themeModule)
    }),
  ])

  return {
    highlighter: currentState.highlighter,
    loadedLanguages: new Set([...currentState.loadedLanguages, ...newLangs]),
    loadedThemes: new Set([...currentState.loadedThemes, ...themes]),
  }
}

export async function createHighlighter({ langs, themes }: { langs: string[], themes: string[] }): Promise<Awaited<ReturnType<typeof createHighlighterCore>>> {
  if (!langs?.length || !themes?.length) {
    throw new Error('Please provide both `langs` and `themes` when creating a highlighter.')
  }

  try {
    const currentState = getState()

    if (currentState) {
      const newState = await loadAdditionalResources(currentState, langs, themes)
      setState(newState)
      return newState.highlighter
    }

    const highlighter = await initializeHighlighter(langs, themes)
    const newState: HighlighterState = {
      highlighter,
      loadedLanguages: new Set(langs),
      loadedThemes: new Set(themes),
    }
    setState(newState)
    return highlighter
  }
  catch (error) {
    console.error('Failed to create highlighter:', error)
    throw error
  }
}

export async function loadTheme(theme: string): Promise<void> {
  const currentState = getState()
  if (!currentState) {
    throw new Error('Highlighter not initialized')
  }

  if (currentState.loadedThemes.has(theme)) {
    return
  }

  const themeModule = await importTheme(theme)
  await currentState.highlighter.loadTheme(themeModule)

  setState({
    ...currentState,
    loadedThemes: new Set([...currentState.loadedThemes, theme]),
  })
}

export async function loadLanguage(lang: string): Promise<void> {
  const currentState = getState()
  if (!currentState) {
    throw new Error('Highlighter not initialized')
  }

  if (currentState.loadedLanguages.has(lang)) {
    return
  }

  const langModule = await importLanguage(lang)
  await currentState.highlighter.loadLanguage(langModule)

  setState({
    ...currentState,
    loadedLanguages: new Set([...currentState.loadedLanguages, lang]),
  })
}

export function processTokens(tokens: ThemedToken[]): TokenType[] {
  return tokens.map(token => ({
    content: token.content,
    color: token.color || '#000000',
    fontStyle: token.fontStyle?.toString(),
  }))
}
